'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var createDefaultJsExpressionContext = function (scope) {
    return {
        getValue: function (variableOrFunctionName) { return scope[variableOrFunctionName]; },
        accessProperty: function (on, propertyName) {
            return on && on.hasOwnProperty(propertyName) ? on[propertyName] : undefined;
        },
    };
};

var isJsExpressionWrappedValue = function (obj) {
    return obj && obj.hasOwnProperty("wrapped") && obj.hasOwnProperty("accessProperty");
};

/* tslint:disable function-constructor no-function-constructor-with-string-args */
var isWord = function (char) {
    return /^[a-zA-Z_]$/.test(char);
};
var isDigit = function (char) {
    return /^[0-9]$/.test(char);
};
var isExpression = function (char) {
    return /^[?:+\-*/()%^=<>!&|, ]$/.test(char);
};
var compileJsExpression = function (expression) {
    var mode = "clean";
    var safeExpression = "return ";
    var invalidChar = function (char) {
        throw new Error("Invalid char at: ".concat(safeExpression).concat(char));
    };
    for (var _i = 0, expression_1 = expression; _i < expression_1.length; _i++) {
        var char = expression_1[_i];
        var backToClean = mode === "word" ? "')" : "";
        if (mode === "string") {
            if (char === '"') {
                mode = "clean";
            }
            else if (char === "\\") {
                mode = "escapedInString";
            }
            safeExpression += char;
        }
        else if (mode === "escapedInString") {
            safeExpression += char;
            mode = "string";
        }
        else if (char === "]") {
            safeExpression += "".concat(backToClean, ")");
            mode = "cleanOrChain";
        }
        else if (mode === "number") {
            if (char === ".") {
                safeExpression += char;
                mode = "decimalPart";
            }
            else if (isExpression(char)) {
                safeExpression += char;
                mode = "clean";
            }
            else if (isDigit(char)) {
                safeExpression += char;
            }
            else {
                invalidChar(char);
            }
        }
        else if (mode === "decimalPart") {
            if (isExpression(char)) {
                mode = "clean";
            }
            else if (!isDigit(char)) {
                invalidChar(char);
            }
            safeExpression += char;
        }
        else if (char === "." && mode !== "clean") {
            // dots are potentially dangerous, they always need to be replaced
            safeExpression += "".concat(backToClean, ".accessProperty('");
            mode = "word";
        }
        else if (char === "[") {
            // another source of danger, replace these as well
            if (mode === "clean") {
                safeExpression += "createArray(";
            }
            else {
                safeExpression += "".concat(backToClean, ".accessProperty(");
                mode = "clean";
            }
        }
        else if (char === "/") {
            safeExpression += "".concat(backToClean, " /1/ "); // Make sure / cannot be used to form a regex, just to do division
            mode = "clean";
        }
        else if (isExpression(char)) {
            safeExpression += backToClean;
            safeExpression += char;
            if (char === ")") {
                mode = "cleanOrChain";
            }
            else {
                mode = "clean";
            }
        }
        else if (mode === "word") {
            if (isWord(char) || isDigit(char)) {
                safeExpression += char;
            }
            else {
                invalidChar(char);
            }
        }
        else {
            // mode === 'clean' || mode === 'cleanOrChain'
            if (isWord(char)) {
                safeExpression += "interpret('".concat(char);
                mode = "word";
            }
            else if (isDigit(char)) {
                safeExpression += char;
                mode = "number";
            }
            else if (char === '"') {
                safeExpression += char;
                mode = "string";
            }
            else {
                invalidChar(char);
            }
        }
    }
    if (mode === "word") {
        safeExpression += "')";
    }
    else if (mode !== "clean" &&
        mode !== "cleanOrChain" &&
        mode !== "number" &&
        mode !== "decimalPart") {
        throw new Error("Invalid end of expression: ".concat(expression));
    }
    var executeFunction = new Function("interpret", "createArray", safeExpression);
    return function (context) {
        var wrapPropertyAccessors = function (wrapped) {
            if (typeof wrapped === "string" ||
                typeof wrapped === "function" ||
                typeof wrapped === "boolean" ||
                typeof wrapped === "number" ||
                wrapped === undefined ||
                wrapped === null) {
                return wrapped;
            }
            else {
                return {
                    wrapped: wrapped,
                    accessProperty: function (propertyName) {
                        return wrapPropertyAccessors(context.accessProperty(wrapped, propertyName));
                    },
                };
            }
        };
        var result = executeFunction(interpret, createArray);
        if (isJsExpressionWrappedValue(result)) {
            return result.wrapped;
        }
        return result;
        function interpret(variableOrFunction) {
            var reservedWordIndex = reservedWords.indexOf(variableOrFunction);
            if (reservedWordIndex !== -1) {
                return reservedValues[reservedWordIndex];
            }
            var value = context.getValue(variableOrFunction);
            if (typeof value === "function") {
                return function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    return wrapPropertyAccessors(value.apply(void 0, args));
                };
            }
            else {
                return wrapPropertyAccessors(value);
            }
        }
        function createArray() {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return {
                wrapped: args.map(function (arg) { return (isJsExpressionWrappedValue(arg) ? arg.wrapped : arg); }),
                accessProperty: function (index) {
                    if (typeof index === "number") {
                        return args[index];
                    }
                },
            };
        }
    };
};
var reservedWords = ["true", "false", "null", "undefined"];
var reservedValues = [true, false, null, undefined];

exports.compileJsExpression = compileJsExpression;
exports.createDefaultJsExpressionContext = createDefaultJsExpressionContext;
exports.isJsExpressionWrappedValue = isJsExpressionWrappedValue;
